#include <stdio.h>
#include "md2.h"

MD2Model::MD2Model()
{
	memset(this,0,sizeof(MD2Model));
}

int MD2Model::load(const char *filename)
{
	FILE *file=fopen(filename,"rb");
	
	if(!file) return 0;

	MD2Header header;
	fread(&header,sizeof(header),1,file);
	
	// Copy the header attributes.
	frameCount=header.numFrames;
	glcmdCount=header.numGlcmds;
	vertexCount=header.numVertices;
	skinCount=header.numSkins;
	texCoordCount=header.numSt;
	triangleCount=header.numTris;
	
	skinWidth=header.skinwidth;
	skinHeight=header.skinheight;
	
	fseek(file,header.offsetSkins,SEEK_SET);
	skins=(MD2Skin *)malloc(sizeof(MD2Skin)*header.numSkins);
	fread(skins,sizeof(MD2Skin),header.numSkins,file);
	
	fseek(file,header.offsetSt,SEEK_SET);
	texCoords=(MD2TexCoord *)malloc(sizeof(MD2TexCoord)*header.numSt);
	fread(texCoords,sizeof(MD2TexCoord),header.numSt,file);
	
	fseek(file,header.offsetTris,SEEK_SET);
	triangles=(MD2Triangle *)malloc(sizeof(MD2Triangle)*header.numTris);
	fread(triangles,sizeof(MD2Triangle),header.numTris,file);
	
	fseek(file,header.offsetFrames,SEEK_SET);
	frames=(MD2Frame *)malloc(sizeof(MD2Frame)*header.numFrames);
	int i;
	for(i=0;i<frameCount;i++) {
		fread(frames+i,sizeof(MD2Frame)-4,1,file);
		frames[i].verts=(MD2Vertex *)malloc(sizeof(MD2Vertex)*vertexCount);
		fread(frames[i].verts,sizeof(MD2Vertex),vertexCount,file);
	}
	
	fseek(file,header.offsetGlcmds,SEEK_SET);
	glcmds=(int *)malloc(sizeof(int)*header.numGlcmds);
	fread(glcmds,sizeof(int),header.numGlcmds,file);
	
	fclose(file);
	
	return 1;
}

Mesh *MD2Model::getFrame(int frame)
{
	if(!mesh) {
		mesh=new Mesh();
		//polyCount=0;	-- done by constructor.
		mesh->vertCount=triangleCount*3;
		mesh->vert=(Vertex3D *)calloc(triangleCount*3,sizeof(Vertex3D));
		mesh->mat.push_back(Material());
		mesh->mat[0].texture=texture;
		mesh->mat[0].textureName=skins[0].name;
	}
	
	MD2Frame *f=frames+frame;
	int i;
	for(i=0;i<triangleCount;i++) {
		// Decompress a triangle from the frame.
		int j;
		for(j=0;j<3;j++) {
			int v0=triangles[i].vertex[j];
			mesh->vert[i*3+j].x=f->verts[v0].v[0]*f->scale[0] + f->translate[0];
			mesh->vert[i*3+j].y=-(f->verts[v0].v[2]*f->scale[2] + f->translate[2]);
			mesh->vert[i*3+j].z=f->verts[v0].v[1]*f->scale[1] + f->translate[1];
			mesh->vert[i*3+j].u=texCoords[triangles[i].st[j]].s/(float)skinWidth;
			mesh->vert[i*3+j].v=texCoords[triangles[i].st[j]].t/(float)skinHeight;
		}
#if 0
		// Calculate a normal over and over. :-)
		float a[3],b[3];
		a[0]=mesh->vert[i*3+1].x-mesh->vert[i*3].x;
		b[0]=mesh->vert[i*3+2].x-mesh->vert[i*3].x;
		a[1]=mesh->vert[i*3+1].y-mesh->vert[i*3].y;
		b[1]=mesh->vert[i*3+2].y-mesh->vert[i*3].y;
		a[2]=mesh->vert[i*3+1].z-mesh->vert[i*3].z;
		b[2]=mesh->vert[i*3+2].z-mesh->vert[i*3].z;
		float nx,ny,nz;
		nx=a[1]*b[2]-a[2]*b[1];
		ny=a[2]*b[0]-a[0]*b[2];
		nz=a[0]*b[1]-a[1]*b[0];
		// Normalize size
		float nw=sqrtf(nx*nx+ny*ny+nz*nz);
		nw=1/nw;
		for(j=0;j<3;j++) {
			mesh->vert[i*3+j].nx=nx*nw;
			mesh->vert[i*3+j].ny=ny*nw;
			mesh->vert[i*3+j].nz=nz*nw;
		}
#else
		// Use magic normals table:
		const struct MD2Normal {
			float nx,ny,nz;
		} normal[]={
			{ -0.525731f,  0.000000f,  0.850651f }, 
			{ -0.442863f,  0.238856f,  0.864188f }, 
			{ -0.295242f,  0.000000f,  0.955423f }, 
			{ -0.309017f,  0.500000f,  0.809017f }, 
			{ -0.162460f,  0.262866f,  0.951056f }, 
			{  0.000000f,  0.000000f,  1.000000f }, 
			{  0.000000f,  0.850651f,  0.525731f }, 
			{ -0.147621f,  0.716567f,  0.681718f }, 
			{  0.147621f,  0.716567f,  0.681718f }, 
			{  0.000000f,  0.525731f,  0.850651f }, 
			{  0.309017f,  0.500000f,  0.809017f }, 
			{  0.525731f,  0.000000f,  0.850651f }, 
			{  0.295242f,  0.000000f,  0.955423f }, 
			{  0.442863f,  0.238856f,  0.864188f }, 
			{  0.162460f,  0.262866f,  0.951056f }, 
			{ -0.681718f,  0.147621f,  0.716567f }, 
			{ -0.809017f,  0.309017f,  0.500000f }, 
			{ -0.587785f,  0.425325f,  0.688191f }, 
			{ -0.850651f,  0.525731f,  0.000000f }, 
			{ -0.864188f,  0.442863f,  0.238856f }, 
			{ -0.716567f,  0.681718f,  0.147621f }, 
			{ -0.688191f,  0.587785f,  0.425325f }, 
			{ -0.500000f,  0.809017f,  0.309017f }, 
			{ -0.238856f,  0.864188f,  0.442863f }, 
			{ -0.425325f,  0.688191f,  0.587785f }, 
			{ -0.716567f,  0.681718f, -0.147621f }, 
			{ -0.500000f,  0.809017f, -0.309017f }, 
			{ -0.525731f,  0.850651f,  0.000000f }, 
			{  0.000000f,  0.850651f, -0.525731f }, 
			{ -0.238856f,  0.864188f, -0.442863f }, 
			{  0.000000f,  0.955423f, -0.295242f }, 
			{ -0.262866f,  0.951056f, -0.162460f }, 
			{  0.000000f,  1.000000f,  0.000000f }, 
			{  0.000000f,  0.955423f,  0.295242f }, 
			{ -0.262866f,  0.951056f,  0.162460f }, 
			{  0.238856f,  0.864188f,  0.442863f }, 
			{  0.262866f,  0.951056f,  0.162460f }, 
			{  0.500000f,  0.809017f,  0.309017f }, 
			{  0.238856f,  0.864188f, -0.442863f }, 
			{  0.262866f,  0.951056f, -0.162460f }, 
			{  0.500000f,  0.809017f, -0.309017f }, 
			{  0.850651f,  0.525731f,  0.000000f }, 
			{  0.716567f,  0.681718f,  0.147621f }, 
			{  0.716567f,  0.681718f, -0.147621f }, 
			{  0.525731f,  0.850651f,  0.000000f }, 
			{  0.425325f,  0.688191f,  0.587785f }, 
			{  0.864188f,  0.442863f,  0.238856f }, 
			{  0.688191f,  0.587785f,  0.425325f }, 
			{  0.809017f,  0.309017f,  0.500000f }, 
			{  0.681718f,  0.147621f,  0.716567f }, 
			{  0.587785f,  0.425325f,  0.688191f }, 
			{  0.955423f,  0.295242f,  0.000000f }, 
			{  1.000000f,  0.000000f,  0.000000f }, 
			{  0.951056f,  0.162460f,  0.262866f }, 
			{  0.850651f, -0.525731f,  0.000000f }, 
			{  0.955423f, -0.295242f,  0.000000f }, 
			{  0.864188f, -0.442863f,  0.238856f }, 
			{  0.951056f, -0.162460f,  0.262866f }, 
			{  0.809017f, -0.309017f,  0.500000f }, 
			{  0.681718f, -0.147621f,  0.716567f }, 
			{  0.850651f,  0.000000f,  0.525731f }, 
			{  0.864188f,  0.442863f, -0.238856f }, 
			{  0.809017f,  0.309017f, -0.500000f }, 
			{  0.951056f,  0.162460f, -0.262866f }, 
			{  0.525731f,  0.000000f, -0.850651f }, 
			{  0.681718f,  0.147621f, -0.716567f }, 
			{  0.681718f, -0.147621f, -0.716567f }, 
			{  0.850651f,  0.000000f, -0.525731f }, 
			{  0.809017f, -0.309017f, -0.500000f }, 
			{  0.864188f, -0.442863f, -0.238856f }, 
			{  0.951056f, -0.162460f, -0.262866f }, 
			{  0.147621f,  0.716567f, -0.681718f }, 
			{  0.309017f,  0.500000f, -0.809017f }, 
			{  0.425325f,  0.688191f, -0.587785f }, 
			{  0.442863f,  0.238856f, -0.864188f }, 
			{  0.587785f,  0.425325f, -0.688191f }, 
			{  0.688191f,  0.587785f, -0.425325f }, 
			{ -0.147621f,  0.716567f, -0.681718f }, 
			{ -0.309017f,  0.500000f, -0.809017f }, 
			{  0.000000f,  0.525731f, -0.850651f }, 
			{ -0.525731f,  0.000000f, -0.850651f }, 
			{ -0.442863f,  0.238856f, -0.864188f }, 
			{ -0.295242f,  0.000000f, -0.955423f }, 
			{ -0.162460f,  0.262866f, -0.951056f }, 
			{  0.000000f,  0.000000f, -1.000000f }, 
			{  0.295242f,  0.000000f, -0.955423f }, 
			{  0.162460f,  0.262866f, -0.951056f }, 
			{ -0.442863f, -0.238856f, -0.864188f }, 
			{ -0.309017f, -0.500000f, -0.809017f }, 
			{ -0.162460f, -0.262866f, -0.951056f }, 
			{  0.000000f, -0.850651f, -0.525731f }, 
			{ -0.147621f, -0.716567f, -0.681718f }, 
			{  0.147621f, -0.716567f, -0.681718f }, 
			{  0.000000f, -0.525731f, -0.850651f }, 
			{  0.309017f, -0.500000f, -0.809017f }, 
			{  0.442863f, -0.238856f, -0.864188f }, 
			{  0.162460f, -0.262866f, -0.951056f }, 
			{  0.238856f, -0.864188f, -0.442863f }, 
			{  0.500000f, -0.809017f, -0.309017f }, 
			{  0.425325f, -0.688191f, -0.587785f }, 
			{  0.716567f, -0.681718f, -0.147621f }, 
			{  0.688191f, -0.587785f, -0.425325f }, 
			{  0.587785f, -0.425325f, -0.688191f }, 
			{  0.000000f, -0.955423f, -0.295242f }, 
			{  0.000000f, -1.000000f,  0.000000f }, 
			{  0.262866f, -0.951056f, -0.162460f }, 
			{  0.000000f, -0.850651f,  0.525731f }, 
			{  0.000000f, -0.955423f,  0.295242f }, 
			{  0.238856f, -0.864188f,  0.442863f }, 
			{  0.262866f, -0.951056f,  0.162460f }, 
			{  0.500000f, -0.809017f,  0.309017f }, 
			{  0.716567f, -0.681718f,  0.147621f }, 
			{  0.525731f, -0.850651f,  0.000000f }, 
			{ -0.238856f, -0.864188f, -0.442863f }, 
			{ -0.500000f, -0.809017f, -0.309017f }, 
			{ -0.262866f, -0.951056f, -0.162460f }, 
			{ -0.850651f, -0.525731f,  0.000000f }, 
			{ -0.716567f, -0.681718f, -0.147621f }, 
			{ -0.716567f, -0.681718f,  0.147621f }, 
			{ -0.525731f, -0.850651f,  0.000000f }, 
			{ -0.500000f, -0.809017f,  0.309017f }, 
			{ -0.238856f, -0.864188f,  0.442863f }, 
			{ -0.262866f, -0.951056f,  0.162460f }, 
			{ -0.864188f, -0.442863f,  0.238856f }, 
			{ -0.809017f, -0.309017f,  0.500000f }, 
			{ -0.688191f, -0.587785f,  0.425325f }, 
			{ -0.681718f, -0.147621f,  0.716567f }, 
			{ -0.442863f, -0.238856f,  0.864188f }, 
			{ -0.587785f, -0.425325f,  0.688191f }, 
			{ -0.309017f, -0.500000f,  0.809017f }, 
			{ -0.147621f, -0.716567f,  0.681718f }, 
			{ -0.425325f, -0.688191f,  0.587785f }, 
			{ -0.162460f, -0.262866f,  0.951056f }, 
			{  0.442863f, -0.238856f,  0.864188f }, 
			{  0.162460f, -0.262866f,  0.951056f }, 
			{  0.309017f, -0.500000f,  0.809017f }, 
			{  0.147621f, -0.716567f,  0.681718f }, 
			{  0.000000f, -0.525731f,  0.850651f }, 
			{  0.425325f, -0.688191f,  0.587785f }, 
			{  0.587785f, -0.425325f,  0.688191f }, 
			{  0.688191f, -0.587785f,  0.425325f }, 
			{ -0.955423f,  0.295242f,  0.000000f }, 
			{ -0.951056f,  0.162460f,  0.262866f }, 
			{ -1.000000f,  0.000000f,  0.000000f }, 
			{ -0.850651f,  0.000000f,  0.525731f }, 
			{ -0.955423f, -0.295242f,  0.000000f }, 
			{ -0.951056f, -0.162460f,  0.262866f }, 
			{ -0.864188f,  0.442863f, -0.238856f }, 
			{ -0.951056f,  0.162460f, -0.262866f }, 
			{ -0.809017f,  0.309017f, -0.500000f }, 
			{ -0.864188f, -0.442863f, -0.238856f }, 
			{ -0.951056f, -0.162460f, -0.262866f }, 
			{ -0.809017f, -0.309017f, -0.500000f }, 
			{ -0.681718f,  0.147621f, -0.716567f }, 
			{ -0.681718f, -0.147621f, -0.716567f }, 
			{ -0.850651f,  0.000000f, -0.525731f }, 
			{ -0.688191f,  0.587785f, -0.425325f }, 
			{ -0.587785f,  0.425325f, -0.688191f }, 
			{ -0.425325f,  0.688191f, -0.587785f }, 
			{ -0.425325f, -0.688191f, -0.587785f }, 
			{ -0.587785f, -0.425325f, -0.688191f }, 
			{ -0.688191f, -0.587785f, -0.425325f }			
		};
		for(j=0;j<3;j++) {
			int v0=triangles[i].vertex[j];
			int n=f->verts[v0].normalIndex;
			mesh->vert[i*3+j].nx=normal[n].nx;
			mesh->vert[i*3+j].ny=normal[n].ny;
			mesh->vert[i*3+j].nz=normal[n].nz;
		}
#endif
	}

#ifdef _DEBUG
	static int dumpId=1;
	char fbuf[256];
	sprintf(fbuf,"dump%02d.msh6",dumpId++);
	dumpId=dumpId%4;
	export_tmsh6(fbuf,mesh);
	printf("Dumped %d\n",dumpId-1);
#endif
	return mesh;
}

Mesh *MD2Model::getFrame(int frame,int nextFrame,float fraction)
{
	// quick and dirty
	
	return getFrame(frame);
}
